@@ -1,4 +1,6 @@
 import os
+import time
+import uuid
 import warnings
 from typing import Dict, Iterable, List, Mapping, Optional, Union
 
@@ -11,14 +13,16 @@ from openai.types.chat import ChatCompletionStreamOptionsParam
 import maitai
 from maitai._config import config
 from maitai._evaluator import Evaluator
-from maitai._inference import InferenceException, InferenceWarning
+from maitai._inference import InferenceException, InferenceWarning, MaitaiConnectionError
 from maitai._types import Body, Headers, MaitaiChunk, MaitaiCompletion, Query, ToolFunction
 from maitai._utils import convert_open_ai_chat_completion_chunk, convert_openai_chat_completion, get_chat_completion_params, required_args
+from maitai.tools import Tools
 from maitai_common.utils.proto_utils import openai_messages_to_proto
 from maitai_common.utils.types import EvaluateCallback
 from maitai_gen.chat import ChatCompletionParams, ChatCompletionRequest, ChatMessage, ClientParams, EvaluationContentType
 from maitai_gen.config import InferenceLocations
 from maitai_gen.inference import InferenceStreamResponse
+from maitai_gen.metric import RequestTimingMetric
 
 DEFAULT_MAX_RETRIES = 2
 
@@ -42,6 +46,8 @@ class Maitai:
         http_client: Optional[httpx.Client] = None,
         _strict_response_validation: bool = False,
     ):
+        if not maitai_api_key:
+            maitai_api_key = os.environ.get("MAITAI_API_KEY")
         if maitai_api_key:
             config.initialize(maitai_api_key)
         if api_key:
@@ -57,9 +63,9 @@ class Maitai:
             config.auth_keys.groq_api_key.key_value = groq_api_key
         if anthropic_api_key:
             config.auth_keys.anthropic_api_key.key_value = anthropic_api_key
-        self.client = None
+        self._client = None
         if api_key or openai_api_key or os.environ.get("OPENAI_API_KEY"):
-            self.client = openai.OpenAI(
+            self._client = openai.OpenAI(
                 api_key=api_key or openai_api_key or os.environ.get("OPENAI_API_KEY"),
                 organization=organization,
                 project=project,
@@ -78,7 +84,7 @@ class Maitai:
                 default_headers=default_headers or {},
                 default_query=default_query or {},
             )
-        self.chat = Chat(self.client, client_params)
+        self.chat = Chat(self._client, client_params)
 
 
 class Chat:
@@ -88,15 +94,15 @@ class Chat:
 
 class Completions:
     def __init__(self, client: Optional[openai.Client] = None, client_params: Optional[ClientParams] = None):
-        self.client = client
-        self.client_params = client_params or ClientParams()
+        self._client = client
+        self._client_params = client_params or ClientParams()
 
-    @required_args(["session_id", "action_type", "application_ref_name", "messages"],
-                   ["session_id", "intent", "application_ref_name", "messages"],
-                   ["session_id", "action_type", "application_ref_name", "messages", "model"],
-                   ["session_id", "intent", "application_ref_name", "messages", "model"],
-                   ["session_id", "action_type", "application_ref_name", "messages", "model", "stream"],
-                   ["session_id", "intent", "application_ref_name", "messages", "model", "stream"])
+    @required_args(["action_type", "application_ref_name", "messages"],
+                   ["intent", "application_ref_name", "messages"],
+                   ["action_type", "application_ref_name", "messages", "model"],
+                   ["intent", "application_ref_name", "messages", "model"],
+                   ["action_type", "application_ref_name", "messages", "model", "stream"],
+                   ["intent", "application_ref_name", "messages", "model", "stream"])
     def create(
         self,
         *,
@@ -134,7 +140,7 @@ class Completions:
         stream_options: Union[Optional[openai_chat_types.ChatCompletionStreamOptionsParam], openai.NotGiven] = openai.NOT_GIVEN,
         temperature: Union[Optional[float], openai.NotGiven] = openai.NOT_GIVEN,
         tool_choice: Union[openai_chat_types.ChatCompletionToolChoiceOptionParam, openai.NotGiven] = openai.NOT_GIVEN,
-        tools: Union[Iterable[Union[openai_chat_types.ChatCompletionToolParam, ToolFunction]], openai.NotGiven] = openai.NOT_GIVEN,
+        tools: Union[Tools, Iterable[Union[openai_chat_types.ChatCompletionToolParam, ToolFunction]], openai.NotGiven] = openai.NOT_GIVEN,
         top_logprobs: Union[Optional[int], openai.NotGiven] = openai.NOT_GIVEN,
         top_p: Union[Optional[float], openai.NotGiven] = openai.NOT_GIVEN,
         user: Union[str, openai.NotGiven] = openai.NOT_GIVEN,
@@ -143,12 +149,18 @@ class Completions:
         extra_body: Optional[Body] = None,
         timeout: Union[float, httpx.Timeout, None, openai.NotGiven] = openai.NOT_GIVEN,
     ) -> Union[MaitaiCompletion, Iterable[MaitaiChunk]]:
+        retry_args = locals()
+        timing = RequestTimingMetric(
+            time_request_start=time.time(),
+        )
         if not config.api_key:
             raise ValueError("Maitai API Key has not been set")
         if server_side_inference is False and apply_corrections is True:
             raise ValueError("server_side_inference must be true to apply_corrections")
         if apply_corrections is True and evaluation_enabled is False:
             raise ValueError("evaluations must be enabled to apply_corrections")
+        if not session_id:
+            session_id = str(uuid.uuid4())
         if not intent and action_type:
             intent = action_type
         maitai_config = config.get_application_action_config(application_ref_name, intent)
@@ -184,6 +196,8 @@ class Completions:
             stream_options = ChatCompletionStreamOptionsParam(include_usage=True)
         if context_retrieval_enabled is None:
             context_retrieval_enabled = maitai_config.context_retrieval_enabled
+        if fallback_model is None and maitai_config.fallback_model:
+            fallback_model = maitai_config.fallback_model
         completion_params = get_chat_completion_params(
             messages=messages,
             model=model,
@@ -224,28 +238,44 @@ class Completions:
                 fallback_model=fallback_model,
                 user_id=user_id,
                 assistant=assistant,
-                client_params=self.client_params,
+                client_params=self._client_params,
             )
-
-            response = maitai.Inference.infer(chat_request, callback, response_timeout)
-            if stream:
-                return _process_inference_stream(response)
-            # ChatCompletion only
-            for resp in response:
-                if resp.warning:
-                    warnings.warn(resp.warning, InferenceWarning)
-                if resp.error:
-                    raise InferenceException(resp.error)
-                return MaitaiCompletion(resp.chat_completion_response)
-            raise InferenceException("No data received")
-        else:
-            if self.client is None:
-                self.client = openai.OpenAI(
-                    default_query=self.client_params.default_query or None,
-                    default_headers=self.client_params.default_headers or None,
-                    base_url=self.client_params.base_url or None,
+            try:
+                timing.time_pre_request = time.time()
+                response = maitai.Inference.infer(chat_request, callback, response_timeout)
+                if stream:
+                    if 'self' in retry_args:
+                        del retry_args['self']
+                    retry_args['server_side_inference'] = False
+                    retry_args['model'] = model
+                    retry_args['fallback_model'] = fallback_model
+                    return _process_inference_stream(response, retry_args, self.create, timing)
+                # ChatCompletion only
+                for resp in response:
+                    if resp.warning:
+                        warnings.warn(resp.warning, InferenceWarning)
+                    if resp.error:
+                        raise InferenceException(resp.error)
+                    timing.request_id = resp.chat_completion_response.request_id
+                    timing.time_request_end = time.time()
+                    maitai.Inference.store_request_timing_data(timing)
+                    return MaitaiCompletion(resp.chat_completion_response)
+                raise InferenceException("No data received")
+            except MaitaiConnectionError as e:
+                if "gpt" not in model and fallback_model and "gpt" in fallback_model:
+                    model = fallback_model
+                if "gpt" not in model:
+                    raise e
+                print("Maitai issue, retrying on client side")
+                server_side_inference = False
+        if not server_side_inference:
+            if self._client is None:
+                self._client = openai.OpenAI(
+                    default_query=self._client_params.default_query or None,
+                    default_headers=self._client_params.default_headers or None,
+                    base_url=self._client_params.base_url or None,
                 )
-            response = self.client.chat.completions.create(
+            response = self._client.chat.completions.create(
                 messages=messages,
                 model=model,
                 frequency_penalty=frequency_penalty,
@@ -354,14 +384,32 @@ def _process_openai_stream(session_id: Union[str, int],
         )
 
 
-def _process_inference_stream(stream: Iterable[InferenceStreamResponse]) -> Iterable[MaitaiChunk]:
-    for infer_resp in stream:
-        if infer_resp.warning:
-            warnings.warn(infer_resp.warning, InferenceWarning)
-        if infer_resp.error:
-            raise InferenceException(infer_resp.error)
-        chunk = infer_resp.chat_completion_chunk
-        if chunk is not None:
-            yield MaitaiChunk(chunk)
-            if chunk.choices and chunk.choices[0].finish_reason:
-                return
+def _process_inference_stream(stream: Iterable[InferenceStreamResponse], retry_args: dict, retry_func, timing: RequestTimingMetric) -> Iterable[MaitaiChunk]:
+    first = True
+    try:
+        for infer_resp in stream:
+            if infer_resp.warning:
+                warnings.warn(infer_resp.warning, InferenceWarning)
+            if infer_resp.error:
+                raise InferenceException(infer_resp.error)
+            chunk = infer_resp.chat_completion_chunk
+            if chunk is not None:
+                if first:
+                    timing.time_first_chunk = time.time()
+                    timing.request_id = chunk.request_id
+                    first = False
+                yield MaitaiChunk(chunk)
+                if chunk.choices and chunk.choices[0].finish_reason:
+                    timing.time_request_end = time.time()
+                    maitai.Inference.store_request_timing_data(timing)
+                    return
+    except MaitaiConnectionError as e:
+        if 'gpt' not in retry_args.get('model', '') and 'gpt' in retry_args.get('fallback_model', ''):
+            retry_args['model'] = retry_args['fallback_model']
+        if 'gpt' in retry_args.get('model', ''):
+            print("Maitai issue, retrying on client side")
+            response = retry_func(**retry_args)
+            for resp in response:
+                yield resp
+        else:
+            raise e
